/*
 * term.c
 * terminal handling routines for pow-wow
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>

#ifdef USETERMIO
#  ifdef APOLLO
#    include "/sys5.3/usr/include/sys/termio.h"
#  else
/*
 * including both termio.h and termios.h might be an overkill, and gives
 * many warnings, but seems to be necessary at times. works anyway.
 */
/* necessary? when? */
/*
 * I use AIX, and it needs termio.h, not termios.h as the rest of
 * the world, so it seems we must include both (ach) - Cosmos.
 */
#    include <termios.h>
#    include <termio.h>
#  endif
#else /* not USETERMIO */
#  include <sys/ioctl.h>
#endif

#ifdef BSD_386
#  include <sys/ioctl_compat.h>
#  define O_RAW RAW
#  define O_ECHO ECHO
#  define O_CBREAK CBREAK
#endif

#if defined(TCSETS) || defined(TCSETATTR)
#  ifndef TCSETS		/* cc for HP-UX  SHOULD define this... */
#    define TCSETS TCSETATTR
#    define TCGETS TCGETATTR
#  endif
typedef struct termios termiostruct;
#else
#  define TCSETS TCSETA
#  define TCGETS TCGETA
typedef struct termio termiostruct;
#endif

#ifdef VSUSP
#  define O_SUSP VSUSP
#else
#  ifdef SWTCH
#    define O_SUSP SWTCH
#  else
#    define O_SUSP SUSP
#  endif
#endif

#include "powwow.h"
#include "edit.h"
#include "term.h"

int ioctl();

#ifdef VT100	/* hard-coded vt100 features if no termcap: */

char kpadstart[] = "", kpadend[] = "";
char begoln[] = "\r";
char clreoln[] = "\033[K", clreoscr[] = "\033[J";
char leftcur[] = "\033[D", rightcur[] = "\033[C",  upcur[] = "\033[A";
char modebold[] = "\033[1m", modeblink[] = "\033[5m", modeinv[] = "\033[7m";
char modeuline[] = "\033[4m", modestandon[] = "", modestandoff[] = "";
char modenorm[] = "\033[m", backupmodenorm[];

static int len_begoln = 1;
static int len_leftcur = 3, len_upcur = 3, gotocost = 8;
int metakey = 0, wrapglitch = 1;
#define insertfinish (0)

#else /* not VT100, termcap function declarations */

int tgetent();
int tgetnum();
int tgetflag();
char *tgetstr();
char *tgoto();

/* terminal escape sequences */
char kpadstart[CAPLEN], kpadend[CAPLEN];
char begoln[CAPLEN], clreoln[CAPLEN], clreoscr[CAPLEN];
char leftcur[CAPLEN], rightcur[CAPLEN], upcur[CAPLEN], curgoto[CAPLEN];
char delchar[CAPLEN], inschar[CAPLEN], insstop[CAPLEN];
	
/* attribute changers: */
char modebold[CAPLEN], modeblink[CAPLEN], modeinv[CAPLEN], modeuline[CAPLEN];
char modestandon[CAPLEN], modestandoff[CAPLEN], modenorm[CAPLEN];
char backupmodenorm[CAPLEN];

static int len_begoln, len_clreoln;
static int len_leftcur, len_upcur, gotocost;
static int deletecost, insertcost, insertfinish;

int metakey;			/* terminal has meta key setting high bit */
int wrapglitch;			/* strange wrap behaviour (a la vt100) */

static int extract();

#endif /* VT100 */

#ifdef USETERMIO
static termiostruct ttybsave;
#else /* not USETERMIO */
static struct sgttyb ttybsave;
static struct tchars tcsave;
static struct ltchars ltcsave;
#endif /* USETERMIO */


/*
 * Terminal handling routines:
 * These are one big mess of left-justified chicken scratches.
 * It should be handled more cleanly...but unix portability is what it is.
 */

/*
 * Set the terminal to character-at-a-time-without-echo mode, and save the
 * original state in ttybsave
 */
void set_terminal()
{
#ifdef USETERMIO
  termiostruct ttyb;
  ioctl(0, TCGETS, &ttyb);
  ttybsave = ttyb;
  ttyb.c_lflag &= ~(ECHO|ICANON);
  ttyb.c_cc[VTIME] = 0;
  ttyb.c_cc[VMIN] = 1;
  /* disable the special handling of the suspend key (handle it ourselves) */
  ttyb.c_cc[O_SUSP] = 0;
  ioctl(0, TCSETS, &ttyb);
#else /* not USETERMIO */
  struct sgttyb ttyb;
  struct ltchars ltc;
  ioctl(0, TIOCGETP, &ttybsave);
  ioctl(0, TIOCGETC, &tcsave);
  ioctl(0, TIOCGLTC, &ltcsave);
  ttyb = ttybsave;
  ttyb.sg_flags = (ttyb.sg_flags|O_CBREAK) & ~O_ECHO;
  ioctl(0, TIOCSETP, &ttyb);
  ltc = ltcsave;
  ltc.t_suspc = -1;
  ioctl(0, TIOCSLTC, &ltc);
#endif
  fputs(kpadstart, stdout);
}

/*
 * Reset the terminal to its original state
 */
void reset_terminal()
{
#ifdef USETERMIO
  ioctl(0, TCSETS, &ttybsave);
#else  /* not USETERMIO */
  ioctl(0, TIOCSETP, &ttybsave);
  ioctl(0, TIOCSETC, &tcsave);
  ioctl(0, TIOCSLTC, &ltcsave);
#endif /* USETERMIO */
  fputs(kpadend, stdout);
  fflush(stdout);
}

/*
 * enable/disable special keys depending on the current linemode
 */
void set_spec_keys()
{
#ifdef USETERMIO
  int i;
  termiostruct ttyb;
  ioctl(0, TCGETS, &ttyb);
  if(linemode & LM_CHAR)  {
     /* char-by-char mode: disable all special keys and set raw mode */
     for(i = 0; i < NCCS; i++)
     	ttyb.c_cc[i] = 0;
     ttyb.c_oflag &= ~OPOST;
  } else {
     /* line at a time mode: enable them, except suspend */
     for(i = 0; i < NCCS; i++)
     	ttyb.c_cc[i] = ttybsave.c_cc[i];
     /* disable the suspend key (handle it ourselves) */
     ttyb.c_cc[O_SUSP] = 0;
     /* set cooked mode */
     ttyb.c_oflag |= OPOST;
  }
  ioctl(0, TCSETS, &ttyb);
#else /* not USETERMIO */
  struct tchars tc = {-1, -1, -1, -1, -1, -1};
  struct ltchars ltc = {-1, -1, -1, -1, -1, -1};
  struct sgttyb ttyb;
  ioctl(0, TIOCGETP, &ttyb);
  if(linemode & LM_CHAR) {
     /* char-by-char mode: set RAW mode*/
     ttyb.sg_flags |= RAW;
  } else {
     /* line-at-a-time mode: enable spec keys, disable RAW */
     tc = tcsave;
     ltc = ltcsave;
     ltc.t_suspc = -1;	/* suspend key remains disabled */
     ttyb.sg_flags &= ~RAW;
  }
  ioctl(0, TIOCSETP, &ttyb);
  ioctl(0, TIOCSETC, &tc);
  ioctl(0, TIOCSLTC, &ltc);
#endif /* USETERMIO */
}


/*
 * read termcap definitions
 */
void settermcap()
{
#ifndef VT100
   struct tc_init_node {
      char cap[4], *buf;
      int *len, critic;
   };
   static struct tc_init_node tc_init[] = {
	{ "cm", curgoto, 0, 1 },
	{ "ce", clreoln, &len_clreoln, 1 },
	{ "cd", clreoscr, 0, 1 },
	{ "nd", rightcur, 0, 1 },
	{ "le", leftcur, &len_leftcur, 0 },
	{ "up", upcur,   &len_upcur,  0 },
	{ "cr", begoln,  &len_begoln, 0 },
	{ "im", inschar, &insertcost, 0 },
	{ "ic", inschar, &insertcost, 0 },
	{ "ei", insstop, &insertcost, 0 },
	{ "dm", delchar, &deletecost, 0 },
	{ "dc", delchar, &deletecost, 0 },
	{ "ed", delchar, &deletecost, 0 },
	{ "me", modenorm, 0, 0 },
	{ "md", modebold, 0, 0 },
	{ "mb", modeblink, 0, 0 },
	{ "mr", modeinv, 0, 0 },
	{ "us", modeuline, 0, 0 },
	{ "so", modestandon, 0, 0 },
	{ "se", modestandoff, 0, 0 },
	{ "ks", kpadstart, 0, 0 },
	{ "ke", kpadend, 0, 0 },
	{ "", NULL, 0, 0 }
   };
   struct tc_init_node *np;
   char tcbuf[2048];		/* by convention, this is enough */
   int i;
#endif /* not VT100 */
#if !defined(VT100) || defined(TELNETBUG)
   char *term = getenv("TERM");
   if(!term) {
      printf("$TERM not set\n"); 
      exit(1);
   }
#endif
#ifdef VT100
   cols = 80;
# ifdef LINES
   lines = LINES;
# else  /* not LINES */
   lines = 24;
# endif /* not LINES */
#else   /* not VT100 */
   switch(tgetent(tcbuf, term)) {
    case 1:
      break;
    case 0:
      fprintf(stderr,
           "There is no entry for '%s' in the terminal data base.\n", term);
      fprintf(stderr,
           "Please set your $TERM environment variable correctly.\n");
      exit(1);
    default:
      syserr("tgetent");
   }
   for(np = tc_init; np->cap[0]; np++)
     if((i = extract(np->cap, np->buf))) {
	if(np->len) *np->len += i;
     } else if(np->critic) {
	fprintf(stderr,
		"Your terminal is not powerful enough, missing '%s'.\n",
		np->cap);
	exit(1);
     }
   if(!len_begoln)
     strcpy(begoln, "\r"), len_begoln = 1;
   if(!len_leftcur)
     strcpy(leftcur, "\b"), len_leftcur = 1;
   cols = tgetnum("co");
   lines = tgetnum("li");
   wrapglitch = tgetflag("xn");
   metakey = tgetflag("km");
   gotocost = strlen(tgoto(curgoto, cols - 1, lines - 1));
   insertfinish = gotocost + len_clreoln;
#endif  /* not VT100 */
   strcpy(backupmodenorm, modenorm);
#ifdef TELNETBUG
   if(strncmp(term, "vt10", 4) == 0) {
      /* might be NCSA Telnet 2.2 for PC, which doesn't reset colours */
      sprintf(modenorm, "\033[;%c%d;%s%dm", 
	      DEFAULTFG<LOWCOLOURS ? '3' : '9',  DEFAULTFG % LOWCOLOURS,
	      DEFAULTBG<LOWCOLOURS ? "4" : "10", DEFAULTBG % LOWCOLOURS);
   }
#endif /* TELNETBUG */
}

/*
 * add the default keypad bindings to the list
 */
void default_keybindings()
{
    /*
     * Note: termcap doesn't have sequences for the numeric keypad, so we just
     * assume they are the same as for a vt100. They can be redefined
     * at runtime anyway (using #bind)
     */
    add_keynode("kp2", "\033Or", key_run_command, "s");
    add_keynode("kp3", "\033Os", key_run_command, "d");
    add_keynode("kp4", "\033Ot", key_run_command, "w");
    add_keynode("kp5", "\033Ou", key_run_command, "exits");
    add_keynode("kp6", "\033Ov", key_run_command, "e");
    add_keynode("kp8", "\033Ox", key_run_command, "n");
    add_keynode("kp9", "\033Oy", key_run_command, "u");
}
    
/*
 * initialize the key binding list
 */
void init_keybindings()
{
   struct b_init_node {
      char label[8], seq[4];
      void (*funct)();
   };
   static struct b_init_node b_init[] = {
	{ "lf",        "\n",           enter_line },
	{ "return",    "\r",           enter_line },
	{ "bs",        "\b",           del_char_left },
	{ "del",       "\177",         del_char_left },
	{ "tab",       "\t",           complete_word },
	{ "^A",        "\001",         beginning_of_line },
	{ "^B",        "\002",         prev_char },
	{ "^D",        "\004",         del_char_right },
	{ "^E",        "\005",         end_of_line },
	{ "^F",        "\006",         next_char },
	{ "^K",        "\013",         kill_to_eol },
	{ "^L",        "\014",         redraw_input_line },
	{ "^N",        "\016",         next_line },
	{ "^P",        "\020",         prev_line },
	{ "^T",        "\024",         transpose_chars },
	{ "^W",        "\027",         enter_into_history },
	{ "^Z",        "\032",         suspend_powwow },
	{ "M-tab",     "\033\t",       complete_line },
	{ "M-B",       "\033b",        prev_word },
	{ "M-D",       "\033d",        del_word_right },
	{ "M-F",       "\033f",        next_word },
	{ "M-bs",      "\033\b",       del_word_left },
	{ "M-del",     "\033\177",     del_word_left },
	{ "",          "",             0       }
   };
   struct b_init_node *p = b_init;
   do {
      add_keynode(p->label, p->seq, p->funct, NULL);
   } while((++p)->seq[0]);

#ifdef VT100
   add_keynode("left", "\033[D", prev_char, NULL);
   add_keynode("right", "\033[C", next_char, NULL);
   add_keynode("up", "\033[A", prev_line, NULL);
   add_keynode("down", "\033[B", next_line, NULL);
#else /* not VT100 */
   {
      char buf[CAPLEN];
      if(*buf = 0, extract("kl", buf))
	{ add_keynode("left", buf, prev_char, NULL);  *buf = 0; }
      if(extract("kr", buf))
	{ add_keynode("right", buf, next_char, NULL);  *buf = 0; }
      if(extract("ku", buf))
	{ add_keynode("up", buf, prev_line, NULL);  *buf = 0; }
      if(extract("kd", buf))
	add_keynode("down", buf, next_line, NULL);
   }
#endif /* not VT100 */
}

#ifndef VT100
/*
 * extract termcap 'cap' and strcat it to buf.
 * return the lenght of the extracted string.
 */
static int extract(cap, buf)
char *cap, *buf;
{
   static char *bp;
   char *d = buf + strlen(buf);
   char *s = tgetstr(cap, (bp = d, &bp));
   int len;
   if(!s) return (*bp = 0);
   /*
    * Remove the padding information. We assume that no terminals
    * need padding nowadays. At least it makes things much easier.
    */
   s += strspn(s, "0123456789*");
   for(len = 0; *s; *d++ = *s++, len++)
     if(*s == '$' && *(s + 1) == '<')
       if(!(s = strchr(s, '>')) || !*++s) break;
   *d = 0;
   return len;
}
#endif /* not VT100 */


/*
 * position the cursor using absolute coordinates
 * note: does not flush the output buffer
 */
void gotoxy(col, line)
int col, line;
{
#ifdef VT100
  printf("\033[%d;%dH", line + 1, col + 1);
#else
  fputs(tgoto(curgoto, col, line), stdout);
#endif
}

/*
 * optimized cursor movement
 * from (fromline, fromcol) to (toline, tocol)
 * if tocol > 0, (toline, tocol) must lie on editline.
 */
void optmove(fromline, fromcol, toline, tocol)
int fromline, fromcol, toline, tocol;
{
    static char buf[CAPLEN];
    char *cp = buf;
    int cost, i, dist;
    /* First, move vertically to the correct line, then horizontally
     * to the right column. If this turns out to be fewer characters
     * than a direct cursor positioning (gotoxy), use that.
     */
    for(;;) {	/* gotoless */
	if((i = toline - fromline) < 0) {
	    if(!len_upcur || (cost = -i * len_upcur) >= gotocost) break;
	    do strcpy(cp, upcur), cp += len_upcur; while(++i);
	} else if((cost = 2 * i)) {	/* lf is mapped to crlf on output */
	    if(cost >= gotocost) break;
	    do *cp++ = '\n'; while(--i);
	    fromcol = 0;
	}
	if((i = tocol - fromcol) < 0) {
	    dist = -i * len_leftcur;
	    if(dist <= len_begoln + tocol) {
		if((cost += dist) > gotocost) break;
		do strcpy(cp, leftcur), cp += len_leftcur; while(++i);
	    } else {
		if((cost += len_begoln) > gotocost) break;
		strcpy(cp, begoln);  cp += len_begoln;
		fromcol = 0;  i = tocol;
	    }
	}
	if(i) {
	    /* if hiliting in effect, just use gotoxy */
	    if(cost + i > gotocost || *edattrbeg) break;
	    /* FEATURE: no check is performed whether promptstr consists of
	     *  printable chars only. gotoxy might cost less if it doesn't */
	    if(fromcol < col0 && toline == line0) {
		strcpy(cp, promptstr);
		cp += strlen(cp);
		fromcol = col0;
	    }
	    strncpy(cp, edbuf + (toline - line0) * (cols - 1) + fromcol - col0,
		    tocol - fromcol);
	    cp += tocol - fromcol;
	}
	*cp = 0;
	fputs(buf, stdout);
	return;
    }
    gotoxy(tocol, toline);
}


/*
 * clear current input line (prompt == 1 if to clear also prompt)
 * cursor is left right after the prompt
 */
void clearinpline(prompt)
int prompt;
{
    int newcol = prompt ? 0 : col0;
    if(edlen || prompt) {
	optmove(CURLINE(pos), CURCOL(pos), line0, newcol);
	if(line0 < lines - 1)
	    fputs(clreoscr, stdout);
        else
	    fputs(clreoln, stdout);
    }
    col0 = newcol;
}

/*
 * Redraw the input line and put the cursor at the current position.
 * The cursor is assumed to be directly after the prompt.
 */
void redraw_input()
{
    int i, oldline0;

    printf("%s", edattrbeg);

    if(edlen) {
	oldline0 = line0;
	if(edlen < cols - col0 - 1) {
	    printf("%s", edbuf);
	} else {
	    printf("%.*s", cols - col0 - 1, edbuf);
	    for(i = cols - col0 - 1; i <= edlen; i += cols - 1)
		printf("\n%.*s", cols - 1, edbuf + i);
	}
	line0 = lines - (edlen + col0) / (cols - 1) - 1;
	if(oldline0 < line0) line0 = oldline0;
	optmove(CURLINE(edlen), CURCOL(edlen), CURLINE(pos), CURCOL(pos));
    }
}


/*
 * delete n characters at current position (the position is unchanged)
 * assert(n <= edlen - pos)
 */
void deletechars(n)
int n;
{
    int r_cost, d_cost = n, p = pos;
    int nl = p - CURCOL(p);	/* next line's starting pos */
    int cl = CURLINE(p);	/* current line */
    for(d_cost = p + n; d_cost < edlen; d_cost++)
	edbuf[d_cost - n] = edbuf[d_cost];
    for(d_cost -= n; d_cost < edlen; d_cost++)
	edbuf[d_cost] = ' ';
    for( ; ; printf("\n"), p = nl, cl++) {
	d_cost = 0;
	/* FEATURE: ought to be "d_cost = n > gotocost ? -gotocost : -n;"
	   since redraw will need to goto back. Of little importance */
	if((r_cost = edlen) > (nl += cols - 1))
	    r_cost = nl, d_cost = n + gotocost;
	r_cost -= p;
#ifndef VT100
	/* FEATURE: no clreoln is used (it might cost less in the
	  occasion we delete more than one char). Simplicity */
	if(deletecost && deletecost * n + d_cost < r_cost) {
	    for(d_cost = n; d_cost; d_cost--) fputs(delchar, stdout);
	    if(edlen <= nl) break;
	    gotoxy(cols - 1 - n, cl);
	    printf("%.*s", n, edbuf + nl - n - 1);
	} else
#endif
	{
	    printf("%.*s", r_cost, edbuf + p);  p += r_cost;
	    if(edlen <= nl) break;
	}
    }
    edbuf[edlen -= n] = 0;
    switch(pos - p) {
	default: optmove(cl, CURCOL(p), CURLINE(pos), CURCOL(pos)); break;
	case 1: fputs(leftcur, stdout); break;
	case 0: break;
    }
}

/*
 * insert n characters at current position. The position is set to after
 * the inserted characters.
 */
void insertchars(str, n)
char *str; int n;
{
    int r_cost, i_cost, p = pos;
    int nl = p - CURCOL(p);	/* next line's starting pos */
    int cl = CURLINE(p);	/* current line */

    my_memmove(edbuf + p + n, edbuf + p, edlen + 1 - p);
    my_memmove(edbuf + p, str, n);

    edlen += n;  pos += n;
    do {
	i_cost = n;
	if((r_cost = edlen) > (nl += cols - 1))
	    r_cost = nl, i_cost += insertfinish;
	r_cost -= p;
#ifndef VT100
	/* FEATURE: insert mode is used only when one char is inserted
	   (which is probaly true > 95% of the time). Simplicity */
	if(n == 1 && insertcost && insertcost + i_cost < r_cost) {
	    printf("%s%c%s", inschar, edbuf[p++], insstop);
	    if(edlen > nl) {
		gotoxy(cols - 1, cl);
		fputs(clreoln, stdout);
	    }
	} else
#endif /* not VT100 */
	{
	   printf("%.*s", r_cost, edbuf + p);  p += r_cost;
	}
	if(edlen < nl) break;
	printf("\n");  p = nl;
	if(++cl > lines - 1) cl = lines - 1, line0--;
    } while(edlen > nl);
    if(p != pos) optmove(cl, CURCOL(p), CURLINE(pos), CURCOL(pos));
}

/*
 * window size change signal handler:
 * we assume that the bottom line is always the bottom line
 */
void winchsig()
{
    struct winsize wsiz;
  
    /* if ioctl fails or gives silly values, don't change anything */
    if(ioctl(0, TIOCGWINSZ, &wsiz) == 0 && wsiz.ws_row && wsiz.ws_col) {
       lines = wsiz.ws_row;
       cols = wsiz.ws_col;
       send_term_size();
    }
    /* reinstall signal handler for system V unices */
    signal(SIGWINCH, winchsig);
}

